""" Parts of the U-Net model """

import torch
import torch.nn as nn
import torch.nn.functional as F


class DoubleConv(nn.Module):
    """(convolution => [BN] => ReLU) * 2"""

    def __init__(self, in_channels, out_channels, kernel_size, mid_channels=None):
        super().__init__()
        if not mid_channels:
            mid_channels = out_channels
        self.double_conv = nn.Sequential(
            nn.Conv2d(in_channels, mid_channels, kernel_size=kernel_size,
                      padding="same", bias=False),
            nn.BatchNorm2d(mid_channels),
            nn.ReLU(inplace=True),
            nn.Conv2d(mid_channels, out_channels, kernel_size=kernel_size,
                      padding="same", bias=False),
            nn.BatchNorm2d(out_channels),
            nn.ReLU(inplace=True)
        )

    def forward(self, x):
        return self.double_conv(x)


class Down(nn.Module):
    """Downscaling with maxpool then double conv"""

    def __init__(self, in_channels, out_channels, kernel_size):
        super().__init__()
        self.maxpool_conv = nn.Sequential(
            nn.MaxPool2d(2),
            DoubleConv(in_channels, out_channels, kernel_size)
        )

    def forward(self, x):
        return self.maxpool_conv(x)


class Up(nn.Module):
    """Upscaling then double conv"""

    def __init__(self, in_channels, out_channels, kernel_size, bilinear=True):
        super().__init__()

        # if bilinear, use the normal convolutions to reduce the number of channels
        if bilinear:
            self.up = nn.Upsample(scale_factor=2, mode='bilinear', align_corners=True)
            self.conv = DoubleConv(in_channels, out_channels,
                                   kernel_size, in_channels // 2)
        else:
            self.up = nn.ConvTranspose2d(in_channels, in_channels // 2, kernel_size=2, stride=2)
            self.conv = DoubleConv(in_channels, out_channels, kernel_size)

    def forward(self, x1, x2):
        x1 = self.up(x1)
        diffY = x2.size()[2] - x1.size()[2]
        diffX = x2.size()[3] - x1.size()[3]

        x1 = F.pad(x1, [diffX // 2, diffX - diffX // 2,
                        diffY // 2, diffY - diffY // 2])
        # if you have padding issues, see
        # https://github.com/HaiyongJiang/U-Net-Pytorch-Unstructured-Buggy/commit/0e854509c2cea854e247a9c615f175f76fbb2e3a
        # https://github.com/xiaopeng-liao/Pytorch-UNet/commit/8ebac70e633bac59fc22bb5195e513d5832fb3bd
        x = torch.cat([x2, x1], dim=1)
        return self.conv(x)


class OutConv(nn.Module):
    def __init__(self, in_channels, out_channels):
        super(OutConv, self).__init__()
        self.conv = nn.Conv2d(in_channels, out_channels, kernel_size=1)
        self.sigmoid = nn.Sigmoid()

    def forward(self, x):
        x = self.conv(x)
        x = self.sigmoid(x)
        return x


class UNet(nn.Module):
    def __init__(self, n_channels, days_out=1, bilinear=False, kernel_size=3):
        super(UNet, self).__init__()
        self.n_channels = n_channels
        #self.n_classes = n_classes
        self.bilinear = bilinear

        self.inc = (DoubleConv(n_channels, 64, kernel_size))
        self.down1 = (Down(64, 128, kernel_size))
        self.down2 = (Down(128, 256, kernel_size))
        self.down3 = (Down(256, 512, kernel_size))
        factor = 2 if bilinear else 1
        self.down4 = (Down(512, 1024 // factor, kernel_size))
        self.up1 = (Up(1024, 512 // factor, kernel_size, bilinear))
        self.up2 = (Up(512, 256 // factor, kernel_size, bilinear))
        self.up3 = (Up(256, 128 // factor, kernel_size, bilinear))
        self.up4 = (Up(128, 64, kernel_size, bilinear))
        #self.outc = (OutConv(64, n_classes))
        self.outc = (OutConv(64, days_out))

    def forward(self, x):
        #print("x shape 1: ", x.shape)
        x1 = self.inc(x)
        #print("x shape 2: ", x1.shape)
        x2 = self.down1(x1)
        #print("x shape 3: ", x2.shape)
        x3 = self.down2(x2)
        #print("x shape 4: ", x3.shape)
        x4 = self.down3(x3)
        #print("x shape 5: ", x4.shape)
        x5 = self.down4(x4)
        #print("x shape 6: ", x5.shape)
        x = self.up1(x5, x4)
        #print("x shape 7: ", x.shape)
        x = self.up2(x, x3)
        #print("x shape 8: ", x.shape)
        x = self.up3(x, x2)
        #print("x shape 9: ", x.shape)
        x = self.up4(x, x1)
        #print("x shape 10: ", x.shape)
        logits = self.outc(x)
        #print("x shape 11: ", logits.shape)

        return logits

    def use_checkpointing(self):
        self.inc = torch.utils.checkpoint(self.inc)
        self.down1 = torch.utils.checkpoint(self.down1)
        self.down2 = torch.utils.checkpoint(self.down2)
        self.down3 = torch.utils.checkpoint(self.down3)
        self.down4 = torch.utils.checkpoint(self.down4)
        self.up1 = torch.utils.checkpoint(self.up1)
        self.up2 = torch.utils.checkpoint(self.up2)
        self.up3 = torch.utils.checkpoint(self.up3)
        self.up4 = torch.utils.checkpoint(self.up4)
        self.outc = torch.utils.checkpoint(self.outc)